# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)
# @author Simon Heybrock
import pytest
import scipp as sc
from scipp.testing import assert_identical

from scippneutron.tof import chopper_cascade


def test_subframe_init_raises_if_time_and_wavelength_have_different_dims() -> None:
    time = sc.array(dims=['vertex'], values=[0.0, 1.0, 1.0, 0.0], unit='s')
    wavelength = sc.array(dims=['vertex'], values=[1.0, 1.0, 2.0, 2.0], unit='angstrom')
    with pytest.raises(sc.DimensionError):
        chopper_cascade.Subframe(time=time, wavelength=wavelength[0:3])
    with pytest.raises(sc.DimensionError):
        chopper_cascade.Subframe(
            time=time, wavelength=wavelength.rename_dims({'vertex': 'point'})
        )


def test_subframe_init_raises_if_time_cannot_be_converted_to_seconds() -> None:
    time = sc.array(dims=['vertex'], values=[0.0, 1.0, 1.0, 0.0], unit='m')
    wavelength = sc.array(dims=['vertex'], values=[1.0, 1.0, 2.0, 2.0], unit='angstrom')
    with pytest.raises(sc.UnitError):
        chopper_cascade.Subframe(time=time, wavelength=wavelength)


def test_subframe_init_raises_if_wavelength_cannot_be_converted_to_angstrom() -> None:
    time = sc.array(dims=['vertex'], values=[0.0, 1.0, 1.0, 0.0], unit='s')
    wavelength = sc.array(dims=['vertex'], values=[1.0, 1.0, 2.0, 2.0], unit='s')
    with pytest.raises(sc.UnitError):
        chopper_cascade.Subframe(time=time, wavelength=wavelength)


def test_subframe_is_regular() -> None:
    # Triangle with last point after base
    time = sc.array(dims=['vertex'], values=[0.0, 2.0, 3.0], unit='s')
    wavelength = sc.array(dims=['vertex'], values=[1.0, 1.0, 2.0], unit='angstrom')
    subframe = chopper_cascade.Subframe(time=time, wavelength=wavelength)
    assert subframe.is_regular()
    # Triangle with last point inside base
    time = sc.array(dims=['vertex'], values=[0.0, 2.0, 1.0], unit='s')
    subframe = chopper_cascade.Subframe(time=time, wavelength=wavelength)
    assert not subframe.is_regular()
    # Triangle standing on its tip, to also test the min-wavelength
    time = sc.array(dims=['vertex'], values=[1.0, 0.0, 3.0], unit='s')
    wavelength = sc.array(dims=['vertex'], values=[2.0, 1.0, 2.0], unit='angstrom')
    subframe = chopper_cascade.Subframe(time=time, wavelength=wavelength)
    assert subframe.is_regular()
    time = sc.array(dims=['vertex'], values=[1.0, 2.0, 3.0], unit='s')
    subframe = chopper_cascade.Subframe(time=time, wavelength=wavelength)
    assert not subframe.is_regular()


def test_subframe_propagate_by() -> None:
    time = sc.array(dims=['vertex'], values=[0.0, 1.0, 1.0, 0.0], unit='s')
    wavelength = sc.array(dims=['vertex'], values=[1.0, 1.0, 2.0, 2.0], unit='angstrom')
    subframe = chopper_cascade.Subframe(time=time, wavelength=wavelength)
    propagated = subframe.propagate_by(sc.scalar(1.0, unit='m'))
    assert_identical(propagated.wavelength, subframe.wavelength)
    assert (propagated.time > subframe.time).all()
    # Started at same time, but different wavelength
    assert propagated.time[2] > propagated.time[1]
    # Difference should be proportional to wavelength
    dt = propagated.time - subframe.time
    rtol = sc.scalar(1e-12, unit='')
    assert sc.isclose(dt[0], dt[1], atol=sc.scalar(0.0, unit='s'), rtol=rtol)
    assert sc.isclose(dt[2], dt[3], atol=sc.scalar(0.0, unit='s'), rtol=rtol)
    assert sc.allclose(
        (dt[2:3] / dt[1]), sc.scalar(2.0), atol=sc.scalar(0.0), rtol=rtol
    )


def test_subframe_time_is_converted_to_seconds() -> None:
    time = sc.array(dims=['vertex'], values=[0.0, 1.0, 3.0, 2.0], unit='ms')
    wavelength = sc.array(dims=['vertex'], values=[1.0, 1.1, 2.1, 2.0], unit='nm')
    subframe = chopper_cascade.Subframe(time=time, wavelength=wavelength)
    assert_identical(
        subframe.time,
        sc.array(dims=['vertex'], values=[0.0, 0.001, 0.003, 0.002], unit='s'),
    )


def test_subframe_wavelength_is_converted_to_angstrom() -> None:
    time = sc.array(dims=['vertex'], values=[0.0, 1.0, 3.0, 2.0], unit='ms')
    wavelength = sc.array(dims=['vertex'], values=[1.0, 1.1, 2.1, 2.0], unit='nm')
    subframe = chopper_cascade.Subframe(time=time, wavelength=wavelength)
    assert_identical(
        subframe.wavelength,
        sc.array(dims=['vertex'], values=[10.0, 11.0, 21.0, 20.0], unit='angstrom'),
    )


def test_subframe_start_end_properties() -> None:
    time = sc.array(dims=['vertex'], values=[0.0, 1.0, 3.0, 2.0], unit='ms')
    wavelength = sc.array(dims=['vertex'], values=[1.0, 1.1, 2.1, 2.0], unit='nm')
    subframe = chopper_cascade.Subframe(time=time, wavelength=wavelength)
    assert_identical(subframe.start_time, sc.scalar(0.0, unit='s'))
    assert_identical(subframe.end_time, sc.scalar(0.003, unit='s'))
    assert_identical(subframe.start_wavelength, sc.scalar(10.0, unit='angstrom'))
    assert_identical(subframe.end_wavelength, sc.scalar(21.0, unit='angstrom'))


@pytest.fixture
def frame() -> chopper_cascade.Frame:
    time = sc.array(dims=['vertex'], values=[0.0, 1.0, 3.0, 2.0], unit='ms')
    wavelength = sc.array(dims=['vertex'], values=[1.0, 1.1, 2.1, 2.0], unit='nm')
    subframe1 = chopper_cascade.Subframe(time=time, wavelength=wavelength)
    subframe2 = chopper_cascade.Subframe(time=time * 2.0, wavelength=wavelength * 3.0)
    return chopper_cascade.Frame(
        distance=sc.scalar(1.0, unit='m'), subframes=[subframe1, subframe2]
    )


def test_frame_propagate_to(frame: chopper_cascade.Frame) -> None:
    # Propagate to same distance as frame distance, i.e., no propagation
    distance = sc.scalar(1.0, unit='m')
    propagated = frame.propagate_to(distance)
    assert propagated.distance == distance
    assert propagated == frame
    distance = sc.scalar(2.0, unit='m')
    propagated = frame.propagate_to(distance)
    assert propagated.distance == distance
    assert (propagated.subframes[0].time > frame.subframes[0].time).all()
    assert (propagated.subframes[1].time > frame.subframes[1].time).all()
    # Wavelengths are unaffected by propagation
    assert_identical(propagated.subframes[0].wavelength, frame.subframes[0].wavelength)
    assert_identical(propagated.subframes[1].wavelength, frame.subframes[1].wavelength)


def test_frame_propagate_to_works_with_empty_frame() -> None:
    frame = chopper_cascade.Frame(distance=sc.scalar(1.0, unit='m'), subframes=[])
    distance = sc.scalar(2.0, unit='m')
    propagated = frame.propagate_to(distance)
    assert_identical(propagated.distance, distance)
    assert propagated.subframes == []


def test_frame_chop_with_no_effect(frame: chopper_cascade.Frame) -> None:
    # Chopper is open for a long time, it should have no effect
    chopper = chopper_cascade.Chopper(
        distance=frame.distance,
        time_open=sc.array(dims=['slit'], values=[0.0], unit='s'),
        time_close=sc.array(dims=['slit'], values=[1.0], unit='s'),
    )
    assert frame.chop(chopper) == frame


def test_frame_chop_returns_empty_frame_if_chopper_is_closed(
    frame: chopper_cascade.Frame,
) -> None:
    # Chopper is closed during entire arrival time of neutrons
    chopper = chopper_cascade.Chopper(
        distance=frame.distance * 1.0,
        time_open=sc.array(dims=['slit'], values=[1.0], unit='s'),
        time_close=sc.array(dims=['slit'], values=[2.0], unit='s'),
    )
    chopped = frame.chop(chopper)
    assert chopped.distance == chopper.distance
    assert len(chopped.subframes) == 0


def test_frame_chop_works_with_empty_frame() -> None:
    frame = chopper_cascade.Frame(distance=sc.scalar(1.0, unit='m'), subframes=[])
    chopper = chopper_cascade.Chopper(
        distance=frame.distance * 1.5,
        time_open=sc.array(dims=['slit'], values=[0.0], unit='s'),
        time_close=sc.array(dims=['slit'], values=[1.0], unit='s'),
    )
    chopped = frame.chop(chopper)
    assert_identical(chopped.distance, chopper.distance)
    assert chopped.subframes == []


def test_frame_chop_returns_frame_with_distance_set_to_chopper_distance(
    frame: chopper_cascade.Frame,
) -> None:
    chopper = chopper_cascade.Chopper(
        distance=sc.scalar(2.0, unit='m'),
        time_open=sc.array(dims=['slit'], values=[0.0], unit='s'),
        time_close=sc.array(dims=['slit'], values=[1.0], unit='s'),
    )
    chopped = frame.chop(chopper)
    assert_identical(chopped.distance, chopper.distance)


def test_frame_chop_raises_if_chopper_distance_is_less_than_frame_distance(
    frame: chopper_cascade.Frame,
) -> None:
    chopper = chopper_cascade.Chopper(
        distance=sc.scalar(0.9, unit='m'),
        time_open=sc.array(dims=['slit'], values=[0.0], unit='s'),
        time_close=sc.array(dims=['slit'], values=[1.0], unit='s'),
    )
    with pytest.raises(ValueError, match='smaller than frame distance'):
        frame.chop(chopper)


def test_frame_chop_trims_subframes_using_chopper_open() -> None:
    time = sc.array(dims=['vertex'], values=[0.0, 2.0, 4.0, 2.0], unit='s')
    wavelength = sc.array(
        dims=['vertex'], values=[10.0, 10.0, 20.0, 20.0], unit='angstrom'
    )
    subframe = chopper_cascade.Subframe(time=time, wavelength=wavelength)
    frame = chopper_cascade.Frame(
        distance=sc.scalar(1.0, unit='m'), subframes=[subframe, subframe]
    )
    chopper = chopper_cascade.Chopper(
        distance=frame.distance,
        time_open=sc.array(dims=['slit'], values=[1.0], unit='s'),
        time_close=sc.array(dims=['slit'], values=[6.0], unit='s'),
    )
    chopped = frame.chop(chopper)
    expected_time = sc.array(
        dims=['vertex'], values=[1.0, 2.0, 4.0, 2.0, 1.0], unit='s'
    )
    # The 15 is from the cut of the line from (2, 20) to (0, 10) in at t=1
    expected_wavelength = sc.array(
        dims=['vertex'], values=[10.0, 10.0, 20.0, 20.0, 15.0], unit='angstrom'
    )
    expected_subframe = chopper_cascade.Subframe(
        time=expected_time, wavelength=expected_wavelength
    )
    expected = chopper_cascade.Frame(
        distance=frame.distance, subframes=[expected_subframe, expected_subframe]
    )
    assert chopped == expected


def test_frame_chop_trims_subframes_using_chopper_close() -> None:
    time = sc.array(dims=['vertex'], values=[0.0, 2.0, 4.0, 2.0], unit='s')
    wavelength = sc.array(
        dims=['vertex'], values=[10.0, 10.0, 20.0, 20.0], unit='angstrom'
    )
    subframe = chopper_cascade.Subframe(time=time, wavelength=wavelength)
    frame = chopper_cascade.Frame(
        distance=sc.scalar(1.0, unit='m'), subframes=[subframe, subframe]
    )
    chopper = chopper_cascade.Chopper(
        distance=frame.distance,
        time_open=sc.array(dims=['slit'], values=[0.0], unit='s'),
        time_close=sc.array(dims=['slit'], values=[3.0], unit='s'),
    )
    chopped = frame.chop(chopper)
    expected_time = sc.array(
        dims=['vertex'], values=[0.0, 2.0, 3.0, 3.0, 2.0], unit='s'
    )
    # The 15 is from the cut of the line from (2, 10) to (4, 20) in at t=1
    expected_wavelength = sc.array(
        dims=['vertex'], values=[10.0, 10.0, 15.0, 20.0, 20.0], unit='angstrom'
    )
    expected_subframe = chopper_cascade.Subframe(
        time=expected_time, wavelength=expected_wavelength
    )
    expected = chopper_cascade.Frame(
        distance=frame.distance, subframes=[expected_subframe, expected_subframe]
    )
    assert chopped == expected


def test_frame_chop_with_multi_slit_chopper_splits_subframes() -> None:
    time = sc.array(dims=['vertex'], values=[0.0, 2.0, 4.0, 2.0], unit='s')
    wavelength = sc.array(
        dims=['vertex'], values=[10.0, 10.0, 20.0, 20.0], unit='angstrom'
    )
    subframe = chopper_cascade.Subframe(time=time, wavelength=wavelength)
    frame = chopper_cascade.Frame(
        distance=sc.scalar(1.0, unit='m'), subframes=[subframe, subframe]
    )
    chopper = chopper_cascade.Chopper(
        distance=frame.distance,
        time_open=sc.array(dims=['slit'], values=[0.0, 3.0], unit='s'),
        time_close=sc.array(dims=['slit'], values=[1.0, 4.0], unit='s'),
    )
    time_left = sc.array(dims=['vertex'], values=[0.0, 1.0, 1.0], unit='s')
    time_right = sc.array(dims=['vertex'], values=[3.0, 4.0, 3.0], unit='s')
    wavelength_left = sc.array(
        dims=['vertex'], values=[10.0, 10.0, 15.0], unit='angstrom'
    )
    wavelength_right = sc.array(
        dims=['vertex'], values=[15.0, 20.0, 20.0], unit='angstrom'
    )
    expected_subframe_left = chopper_cascade.Subframe(
        time=time_left, wavelength=wavelength_left
    )
    expected_subframe_right = chopper_cascade.Subframe(
        time=time_right, wavelength=wavelength_right
    )
    expected = chopper_cascade.Frame(
        distance=frame.distance,
        subframes=[
            expected_subframe_left,
            expected_subframe_right,
            expected_subframe_left,
            expected_subframe_right,
        ],
    )
    assert frame.chop(chopper) == expected


def test_frame_chop_with_multi_slit_chopper_blind_slit() -> None:
    time = sc.array(dims=['vertex'], values=[0.0, 2.0, 4.0, 2.0], unit='s')
    wavelength = sc.array(
        dims=['vertex'], values=[10.0, 10.0, 20.0, 20.0], unit='angstrom'
    )
    subframe = chopper_cascade.Subframe(time=time, wavelength=wavelength)
    frame = chopper_cascade.Frame(
        distance=sc.scalar(1.0, unit='m'), subframes=[subframe, subframe]
    )
    # The second slit opens after all neutrons have passed, it will have no effect.
    chopper = chopper_cascade.Chopper(
        distance=frame.distance,
        time_open=sc.array(dims=['slit'], values=[0.0, 5.0], unit='s'),
        time_close=sc.array(dims=['slit'], values=[3.0, 6.0], unit='s'),
    )
    chopped = frame.chop(chopper)
    equivalent_single_slit_chopper = chopper_cascade.Chopper(
        distance=frame.distance,
        time_open=sc.array(dims=['slit'], values=[0.0], unit='s'),
        time_close=sc.array(dims=['slit'], values=[3.0], unit='s'),
    )
    expected = frame.chop(equivalent_single_slit_chopper)
    assert chopped == expected


def test_frame_bounds_gives_global_min_and_max() -> None:
    time = sc.array(dims=['vertex'], values=[0.0, 2.0, 4.0, 2.0], unit='s')
    wavelength = sc.array(
        dims=['vertex'], values=[10.0, 10.0, 20.0, 20.0], unit='angstrom'
    )
    subframe1 = chopper_cascade.Subframe(time=time, wavelength=wavelength)
    subframe2 = chopper_cascade.Subframe(time=time * 2.0, wavelength=wavelength * 3.0)
    frame = chopper_cascade.Frame(
        distance=sc.scalar(1.0, unit='m'), subframes=[subframe1, subframe2]
    )
    bounds = frame.bounds()
    assert_identical(
        bounds['time'], sc.array(dims=['bound'], values=[0.0, 8.0], unit='s')
    )
    assert_identical(
        bounds['wavelength'],
        sc.array(dims=['bound'], values=[10.0, 60.0], unit='angstrom'),
    )


def test_frame_sequence_sets_up_rectangular_subframe() -> None:
    frames = chopper_cascade.FrameSequence.from_source_pulse(
        time_min=sc.scalar(0.0, unit='s'),
        time_max=sc.scalar(1.0, unit='s'),
        wavelength_min=sc.scalar(1.0, unit='angstrom'),
        wavelength_max=sc.scalar(2.0, unit='angstrom'),
    )
    expected = chopper_cascade.Frame(
        distance=sc.scalar(0.0, unit='m'),
        subframes=[
            chopper_cascade.Subframe(
                time=sc.array(dims=['vertex'], values=[0.0, 1.0, 1.0, 0.0], unit='s'),
                wavelength=sc.array(
                    dims=['vertex'], values=[1.0, 1.0, 2.0, 2.0], unit='angstrom'
                ),
            )
        ],
    )
    assert len(frames) == 1
    assert frames[0] == expected


@pytest.fixture
def source_frame_sequence() -> chopper_cascade.FrameSequence:
    return chopper_cascade.FrameSequence.from_source_pulse(
        time_min=sc.scalar(0.0, unit='ms'),
        time_max=sc.scalar(1.0, unit='ms'),
        wavelength_min=sc.scalar(1.0, unit='angstrom'),
        wavelength_max=sc.scalar(10.0, unit='angstrom'),
    )


def test_frame_sequence_propagate_to_returns_new_sequence_with_added_propagated_frame(
    source_frame_sequence: chopper_cascade.FrameSequence,
) -> None:
    frames = source_frame_sequence
    distance = sc.scalar(1.5, unit='m')
    result = frames.propagate_to(distance)
    assert len(frames) == 1
    assert len(result) == 2
    assert result[1] == frames[0].propagate_to(distance)
    result2 = result.propagate_to(distance * 2)
    assert len(result2) == 3
    assert result2[2] == frames[0].propagate_to(distance * 2)
    assert result2[2] == result[1].propagate_to(distance * 2)


def test_frame_sequence_propagate_to_array_of_distances(
    source_frame_sequence: chopper_cascade.FrameSequence,
) -> None:
    frames = source_frame_sequence
    distances = sc.array(dims=['distance'], values=[1.5, 1.7], unit='m')
    result = frames.propagate_to(distances)
    assert len(frames) == 1
    assert len(result) == 2
    assert result[1] == frames[0].propagate_to(distances)
    subframe = result[1].subframes[0]
    assert subframe.start_time.sizes["distance"] == 2
    assert subframe.end_time.sizes["distance"] == 2


def test_frame_sequence_propagate_to_array_of_distances_subbounds(
    source_frame_sequence: chopper_cascade.FrameSequence,
) -> None:
    frames = source_frame_sequence
    distances = sc.array(dims=['distance'], values=[1.5, 1.7], unit='m')
    result = frames.propagate_to(distances)
    subbounds = result[-1].subbounds()
    subframe = result[1].subframes[0]
    assert_identical(subbounds['time']['subframe', 0]['bound', 0], subframe.start_time)
    assert_identical(subbounds['time']['subframe', 0]['bound', 1], subframe.end_time)
    assert_identical(
        subbounds['wavelength']['subframe', 0]['bound', 0], subframe.start_wavelength
    )
    assert_identical(
        subbounds['wavelength']['subframe', 0]['bound', 1], subframe.end_wavelength
    )


def test_frame_sequence_chop_returns_new_sequence_with_added_chopped_frames(
    source_frame_sequence: chopper_cascade.FrameSequence,
) -> None:
    frames = source_frame_sequence
    chopper1 = chopper_cascade.Chopper(
        distance=sc.scalar(1.5, unit='m'),
        time_open=sc.array(dims=['slit'], values=[0.0], unit='s'),
        time_close=sc.array(dims=['slit'], values=[0.001], unit='s'),
    )
    chopper2 = chopper_cascade.Chopper(
        distance=sc.scalar(2.5, unit='m'),
        time_open=sc.array(dims=['slit'], values=[0.001], unit='s'),
        time_close=sc.array(dims=['slit'], values=[0.003], unit='s'),
    )
    result = frames.chop([chopper1, chopper2])
    assert len(frames) == 1
    assert len(result) == 3  # source + 2 choppers
    assert len(result[2].subframes) == 1  # something makes it through
    assert result[0] == frames[0]
    assert result[0].chop(chopper1) == result[1]
    assert result[1].chop(chopper2) == result[2]


def test_frame_sequence_chop_applies_choppers_ordered_by_distance(
    source_frame_sequence: chopper_cascade.FrameSequence,
) -> None:
    frames = source_frame_sequence
    chopper1 = chopper_cascade.Chopper(
        distance=sc.scalar(1.5, unit='m'),
        time_open=sc.array(dims=['slit'], values=[0.0], unit='s'),
        time_close=sc.array(dims=['slit'], values=[0.001], unit='s'),
    )
    chopper2 = chopper_cascade.Chopper(
        distance=sc.scalar(2.5, unit='m'),
        time_open=sc.array(dims=['slit'], values=[0.001], unit='s'),
        time_close=sc.array(dims=['slit'], values=[0.003], unit='s'),
    )
    result12 = frames.chop([chopper1, chopper2])
    result21 = frames.chop([chopper2, chopper1])
    assert result12 == result21
    assert result12[2] == result21[2]


def test_frame_sequence_getitem_by_distance_selects_correct_chopper(
    source_frame_sequence: chopper_cascade.FrameSequence,
) -> None:
    frames = source_frame_sequence
    chopper1 = chopper_cascade.Chopper(
        distance=sc.scalar(1.5, unit='m'),
        time_open=sc.array(dims=['slit'], values=[0.0], unit='s'),
        time_close=sc.array(dims=['slit'], values=[0.001], unit='s'),
    )
    chopper2 = chopper_cascade.Chopper(
        distance=sc.scalar(2.5, unit='m'),
        time_open=sc.array(dims=['slit'], values=[0.001], unit='s'),
        time_close=sc.array(dims=['slit'], values=[0.003], unit='s'),
    )
    frames = frames.chop([chopper1, chopper2])
    distance = sc.scalar(1.0, unit='m')
    result = frames[distance]
    assert_identical(result.bounds(), frames[0].propagate_to(distance).bounds())
    distance = sc.scalar(2.0, unit='m')
    result = frames[distance]
    assert_identical(result.bounds(), frames[1].propagate_to(distance).bounds())
    distance = sc.scalar(2.4, unit='m')
    result = frames[distance]
    assert_identical(result.bounds(), frames[1].propagate_to(distance).bounds())
    distance = sc.scalar(3.0, unit='m')
    result = frames[distance]
    assert_identical(result.bounds(), frames[2].propagate_to(distance).bounds())



from dataclasses import dataclass
from typing import Literal

import numpy as np
import scipp as sc
from numpy.polynomial.chebyshev import chebgauss
from numpy.polynomial.legendre import leggauss

from . import quadratures
from .types import SampleShape


@dataclass
class Cylinder(SampleShape):
    symmetry_line: sc.Variable
    center_of_base: sc.Variable
    radius: sc.Variable
    height: sc.Variable

    def beam_intersection(
        self, start_point: sc.Variable, direction: sc.Variable
    ) -> sc.Variable:
        """Length of intersection between beam and cylinder"""
        base_point = self.center_of_base - start_point
        cyl_intersection, *cyl_interval = _line_infinite_cylinder_intersection(
            self.symmetry_line, base_point, self.radius, direction
        )
        slab_intersection, *slab_interval = _line_slab_intersection(
            self.symmetry_line, base_point, self.height, direction
        )
        return sc.where(
            cyl_intersection & slab_intersection,
            _positive_interval_intersection(slab_interval, cyl_interval),
            sc.scalar(0.0, unit=start_point.unit),
        )

    @property
    def center(self) -> sc.Variable:
        return self.center_of_base + self.symmetry_line * self.height / 2

    @property
    def volume(self) -> sc.Variable:
        return self.radius**2 * self.height * np.pi

    def _select_quadrature_points(self, kind):
        if kind == 'expensive':
            k = round(max(min(11 * (self.height / self.radius).value, 35), 11))
            x, w = chebgauss(k)
            w *= (1 - x**2) ** 0.5
            w /= sum(w) / 2
            quad = _cylinder_quadrature_from_product(
                quadratures.disk256_cheb,
                dict(x=x, weights=w),  # noqa: C408
            )
        elif kind == 'medium':
            k = round(max(min(7 * (self.height / self.radius).value, 25), 7))
            x, w = chebgauss(k)
            w *= (1 - x**2) ** 0.5
            w /= sum(w) / 2
            # Would be nice to have a medium size Chebychev quadrature on the disk,
            # but I only found the large one for now.
            quad = _cylinder_quadrature_from_product(
                quadratures.disk55,
                dict(x=x, weights=w),  # noqa: C408
            )
        elif kind == 'cheap':
            k = round(max(min(5 * (self.height / self.radius).value, 15), 5))
            x, w = leggauss(k)
            quad = _cylinder_quadrature_from_product(
                quadratures.disk12,
                dict(x=x, weights=w),  # noqa: C408
            )
        elif (
            kind == 'mc'
            or isinstance(kind, tuple)
            and len(kind) > 0
            and kind[0] == 'mc'
        ):
            # The Monte-Carlo integration option is useful because
            # while inefficient it is guaranteed unbiased and good for debugging.
            k = 5000 if kind == 'mc' or len(kind) == 1 else kind[1]
            # Uniform sampling in cylinder
            r = np.random.random(k) ** 0.5
            th = 2 * np.pi * np.random.random(k)
            z = 2 * np.random.random(k) - 1
            quad = {
                'x': r * np.cos(th),
                'y': r * np.sin(th),
                'z': z,
                'weights': 2 * np.pi * np.ones(k) / k,
            }
        else:
            raise NotImplementedError
        return {k: sc.array(dims=['quad'], values=v) for k, v in quad.items()}

    def quadrature(
        self,
        kind: Literal['expensive', 'medium', 'cheap', 'mc'] | tuple[Literal['mc'], int],
    ) -> tuple[sc.Variable, sc.Variable]:
        'Returns quadrature points and weights of the cylinder.'
        quad = self._select_quadrature_points(kind)
        # Scale to size of cylinder
        x = (quad['x'] * self.radius).to(unit=self.center.unit)
        y = (quad['y'] * self.radius).to(unit=self.center.unit)
        z = (quad['z'] * self.height / 2).to(unit=self.center.unit)
        # Quadrature had total weight 2pi before, now it is 2pi radius^2 * height
        weights = quad['weights'] * (self.radius**2 * self.height / 2)
        points = sc.vectors(
            dims=['quad'],
            values=(sc.concat([x, y, z], dim='row').transpose(['quad', 'row']).values),
            unit=self.center.unit,
        )
        # By default the cylinder quadrature has z as the symmetry axis.
        # We need to rotate the quadrature to match the symmetry axis of the cylinder.
        u = sc.cross(sc.vector([0, 0, 1]), self.symmetry_line)
        un = sc.norm(u)
        if un >= 1e-10:
            u *= sc.asin(un) / un
            points = sc.spatial.rotations_from_rotvecs(u) * points

        # By default the cylinder quadrature center is at the origin.
        # We need to move it so the center matches the cylinder.
        points += self.center
        return points, weights


def _cylinder_quadrature_from_product(disk_quadrature, line_quadrature):
    return dict(  # noqa: C408
        weights=np.array(
            [
                disk_w * line_w
                for disk_w in disk_quadrature['weights']
                for line_w in line_quadrature['weights']
            ]
        ),
        x=np.repeat(disk_quadrature['x'], len(line_quadrature['weights'])),
        y=np.repeat(disk_quadrature['y'], len(line_quadrature['weights'])),
        z=np.tile(line_quadrature['x'], len(disk_quadrature['weights'])),
    )


def _minimum(x, y):
    return sc.where(x <= y, x, y)


def _maximum(x, y):
    return sc.where(x >= y, x, y)


def _max0(x):
    return _maximum(x, sc.scalar(0.0, unit=x.unit))


def _positive_interval_intersection(a, b):
    '''Length of the intersection of a and b and the positive real axis'''
    left, right = _maximum(a[0], b[0]), _minimum(a[1], b[1])
    return _max0(_max0(right) - _max0(left))


def _line_infinite_cylinder_intersection(a, b, r, n):
    '''Intersection between a cylinder and a line through the origin with direction n.
    See https://en.wikipedia.org/wiki/Line-cylinder_intersection.

    Parameters:
    -----------
    a: cylinder symmetry line
    b: point on cylinder symmetry line
    r: radius
    n: direction of line to intersect cylinder

    If the cylinder intersects the line through the origin in the direction of n
    then `intersection` is `True` and the cylinder and the line intersects all
    points `t * n` for `left <= t <= right`.

    Returns:
    ---------
    intersection:
        True if the line and the cylinder intersect anywhere
    left:
        first edge of intersection segment (direction opposite n)
    right:
        second edge of intersection segment (direction n)
    '''
    nxa = sc.cross(n, a)
    nxa_square = sc.dot(nxa, nxa)
    parallel_to_cylinder = nxa_square == sc.scalar(0.0, unit=nxa.unit)
    s2 = nxa_square * r**2 - sc.dot(b, nxa) ** 2
    s = sc.sqrt(s2)
    m = sc.dot(nxa, sc.cross(b, a))
    intersection = s2 >= sc.scalar(0.0, unit=s2.unit)
    left = sc.where(
        parallel_to_cylinder,
        sc.scalar(float('-inf'), unit=m.unit),
        (m - s) / nxa_square,
    )
    right = sc.where(
        parallel_to_cylinder, sc.scalar(float('inf'), unit=m.unit), (m + s) / nxa_square
    )
    origin_in_cylinder = sc.norm(b - sc.dot(b, a) * a) <= r
    return (
        sc.where(parallel_to_cylinder, origin_in_cylinder, intersection),
        left,
        right,
    )


def _line_slab_intersection(a, b, h, n):
    '''Intersection between a slab (the volume between two parallel planes)
    and a line through the origin with direction n.

    Returns:
    --------
    intersection:
        True if the line and the slab intersect anywhere
    left:
        first edge of intersection segment (direction opposite n)
    right:
        second edge of intersection segment (direction n)
    '''
    ndota = sc.dot(n, a)
    bdota = sc.dot(b, a)
    origin_in_plane = (bdota <= sc.scalar(0.0, unit=h.unit)) & (bdota >= -h)
    parallel_to_slab = sc.abs(ndota) == sc.scalar(0, unit=ndota.unit)
    t0 = bdota / ndota
    t1 = t0 + h / ndota
    left = _minimum(t0, t1)
    right = _maximum(t1, t0)
    return (
        origin_in_plane | (~parallel_to_slab),
        sc.where(
            parallel_to_slab,
            sc.scalar(float('-inf'), unit=b.unit),
            left,
        ),
        sc.where(
            parallel_to_slab,
            sc.scalar(float('inf'), unit=b.unit),
            right,
        ),
    )



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2024 Scipp contributors (https://github.com/scipp)

import numpy as np
import pytest
import scipp as sc
import scipp.testing

from scippneutron.peaks import fit_peaks
from scippneutron.peaks.model import GaussianModel, LorentzianModel, PolynomialModel


def test_fit_peaks_rejects_descending_coord():
    data = sc.DataArray(sc.ones(sizes={'x': 10}), coords={'x': -sc.arange('x', 10.0)})
    with pytest.raises(sc.CoordError):
        fit_peaks(
            data,
            peak_estimates=sc.array(dims=['x'], values=[5.0]),
            windows=sc.scalar(1.0),
            background='linear',
            peak='gaussian',
        )


def test_fit_peaks_rejects_disordered_coord():
    rng = np.random.default_rng(93)
    data = sc.DataArray(sc.ones(sizes={'x': 10}), coords={'x': sc.arange('x', 10.0)})
    data = data[list(rng.permutation(len(data)))]
    with pytest.raises(sc.CoordError):
        fit_peaks(
            data,
            peak_estimates=sc.array(dims=['x'], values=[5.0]),
            windows=sc.scalar(1.0),
            background='linear',
            peak='gaussian',
        )


def test_fit_peaks_finds_single_gaussian_peak_linear_background():
    rng = np.random.default_rng(663)
    model = GaussianModel(prefix='peak_') + PolynomialModel(degree=1, prefix='bkg_')
    real_params = {
        'peak_amplitude': sc.scalar(2.7, unit='K*m'),
        'peak_loc': sc.scalar(14.0, unit='m'),
        'peak_scale': sc.scalar(2.8, unit='m'),
        'bkg_a0': sc.scalar(-4.9, unit='K'),
        'bkg_a1': sc.scalar(1.8, unit='K/m'),
    }

    x = sc.linspace('x', 3.0, 20.0, 100, unit='m')
    y = model(x, **real_params)
    y.values += rng.normal(0.0, 0.01, len(y))
    y.variances = y.values / 3.0
    data = sc.DataArray(y, coords={'x': x})

    [result] = fit_peaks(
        data,
        peak_estimates=sc.array(dims=['x'], values=[13.0], unit='m'),
        windows=sc.scalar(15.0, unit='m'),
        background='linear',
        peak='gaussian',
    )
    assert result.success
    sc.testing.assert_allclose(
        sc.values(result.popt['peak_amplitude']),
        real_params['peak_amplitude'],
        rtol=sc.scalar(0.1),
    )
    sc.testing.assert_allclose(
        sc.values(result.popt['peak_loc']), real_params['peak_loc'], rtol=sc.scalar(0.1)
    )
    sc.testing.assert_allclose(
        sc.values(result.popt['peak_scale']),
        real_params['peak_scale'],
        rtol=sc.scalar(0.1),
    )
    sc.testing.assert_allclose(
        sc.values(result.popt['bkg_a0']), real_params['bkg_a0'], rtol=sc.scalar(0.1)
    )
    sc.testing.assert_allclose(
        sc.values(result.popt['bkg_a1']), real_params['bkg_a1'], rtol=sc.scalar(0.1)
    )


def test_fit_peaks_select_model_single_string():
    # The data doesn't really matter here.
    # The test is about ensuring that the fit succeeds without raising.
    x = sc.linspace('x', 0.0, 20.0, 10, unit='m')
    y = sc.scalar(2.3, unit='K') + sc.scalar(-0.1, unit='K/m') * x
    data = sc.DataArray(y, coords={'x': x})
    [result] = fit_peaks(
        data,
        peak_estimates=sc.array(dims=['x'], values=[10.0], unit='m'),
        windows=sc.scalar(15.0, unit='m'),
        background='linear',
        peak='gaussian',
    )
    # There is no peak in the data
    assert not result.success
    assert isinstance(result.background, PolynomialModel)
    assert result.background.degree == 1
    assert isinstance(result.peak, GaussianModel)


def test_fit_peaks_select_model_single_model():
    # The data doesn't really matter here.
    # The test is about ensuring that the fit succeeds without raising.
    x = sc.linspace('x', 0.0, 20.0, 10, unit='m')
    y = sc.scalar(2.3, unit='K') + sc.scalar(-0.1, unit='K/m') * x
    data = sc.DataArray(y, coords={'x': x})
    [result] = fit_peaks(
        data,
        peak_estimates=sc.array(dims=['x'], values=[10.0], unit='m'),
        windows=sc.scalar(15.0, unit='m'),
        background=PolynomialModel(degree=1, prefix='pre_'),
        peak=GaussianModel(prefix='gauss'),
    )
    # There is no peak in the data
    assert not result.success
    assert isinstance(result.background, PolynomialModel)
    assert result.background.degree == 1
    assert result.background.prefix == 'bkg_'
    assert isinstance(result.peak, GaussianModel)
    assert result.peak.prefix == 'peak_'


# Warning about ill-conditioned polynomial.
# This does not matter for the test.
@pytest.mark.filterwarnings('ignore::numpy.polynomial.polyutils.RankWarning')
def test_fit_peaks_select_model_two_strings():
    # The data doesn't really matter here.
    # The test is about ensuring that the fit succeeds without raising.
    x = sc.linspace('x', 0.0, 20.0, 10, unit='m')
    y = sc.scalar(2.3, unit='K') + sc.scalar(-0.1, unit='K/m^2') * x**2
    data = sc.DataArray(y, coords={'x': x})
    [result] = fit_peaks(
        data,
        peak_estimates=sc.array(dims=['x'], values=[10.0], unit='m'),
        windows=sc.scalar(15.0, unit='m'),
        background=('linear', 'quadratic'),
        peak='gaussian',
    )
    # There is no peak in the data
    assert not result.success
    assert isinstance(result.background, PolynomialModel)
    assert result.background.degree in (1, 2)
    assert isinstance(result.peak, GaussianModel)


# Warning about ill-conditioned polynomial.
# This does not matter for the test.
@pytest.mark.filterwarnings('ignore::numpy.polynomial.polyutils.RankWarning')
def test_fit_peaks_select_model_mixed():
    # The data doesn't really matter here.
    # The test is about ensuring that the fit succeeds without raising.
    x = sc.linspace('x', 0.0, 20.0, 10, unit='m')
    y = sc.scalar(2.3, unit='K') + sc.scalar(-0.1, unit='K/m^2') * x**2
    data = sc.DataArray(y, coords={'x': x})
    [result] = fit_peaks(
        data,
        peak_estimates=sc.array(dims=['x'], values=[10.0], unit='m'),
        windows=sc.scalar(15.0, unit='m'),
        background=(PolynomialModel(degree=1, prefix='pre_'), 'quadratic'),
        peak=['lorentzian', GaussianModel(prefix='gauss')],
    )
    # There is no peak in the data
    assert not result.success
    assert isinstance(result.background, PolynomialModel)
    assert result.background.degree in (1, 2)
    assert result.background.prefix == 'bkg_'
    assert isinstance(result.peak, GaussianModel | LorentzianModel)
    assert result.peak.prefix == 'peak_'


def test_fit_peaks_select_model_bad_name():
    # The data doesn't really matter here.
    # The test is about ensuring that the fit succeeds without raising.
    x = sc.linspace('x', 0.0, 20.0, 10, unit='m')
    y = sc.scalar(2.3, unit='K') + sc.scalar(-0.1, unit='K/m^2') * x**2
    data = sc.DataArray(y, coords={'x': x})
    with pytest.raises(ValueError, match='model'):
        fit_peaks(
            data,
            peak_estimates=sc.array(dims=['x'], values=[10.0], unit='m'),
            windows=sc.scalar(15.0, unit='m'),
            background='parabola',
            peak='lorentzian',
        )


def test_fit_peaks_select_model_empty_list():
    # The data doesn't really matter here.
    # The test is about ensuring that the fit succeeds without raising.
    x = sc.linspace('x', 0.0, 20.0, 10, unit='m')
    y = sc.scalar(2.3, unit='K') + sc.scalar(-0.1, unit='K/m^2') * x**2
    data = sc.DataArray(y, coords={'x': x})
    with pytest.raises(ValueError, match='model'):
        fit_peaks(
            data,
            peak_estimates=sc.array(dims=['x'], values=[10.0], unit='m'),
            windows=sc.scalar(15.0, unit='m'),
            background=['linear'],
            peak=[],
        )



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2024 Scipp contributors (https://github.com/scipp)

r"""Tools for disk choppers.

Definitions
-----------

The names used here correspond closely to the names used by NeXus' ``NXdisk_chopper``.
See https://manual.nexusformat.org/classes/base_classes/NXdisk_chopper.html
for an overview.

Here is how those attributes are interpreted in ScippNeutron:
The image below shows a disk chopper with a single slit
as seen from the neutron source looking towards the sample.
Note that all definitions are independent of the rotation direction.

.. list-table::
    :widths: auto
    :header-rows: 1

    * - Name
      - Symbol
      - Definition
    * - ``top_dead_center``
      - :math:`t_0`
      - Choppers typically have a top-dead-center (TDC) sensor that tracks the
        rotation of the chopper.
        The sensor emits a timestamp when a chosen part of the chopper disk passes by.
        This part of the disk serves as a reference point for defining angles;
        it is marked as 'TDC' in the image below.

        The emitted timestamps are stored in the ``top_dead_center`` field
        of a NeXus chopper.
        This serves as a reference time for the chopper :math:`t_0`.
        In :class:`DiskChopper`, the TDC is encoded as a component of
        :attr:`DiskChopper.phase`.
    * - ``beam_position``
      - :math:`\tilde{\theta}`
      - The angle under which the beam hits the chopper
        (:attr:`DiskChopper.beam_position`).
        We do not care about the radial position and assume that the beam can
        pass through all chopper slits.
    * - | ``slit_begin``
        | ``slit_end``
      -
      - Slits are defined in terms of *begin* (:attr:`DiskChopper.slit_begin`,
        :math:`\theta` in the image) and *end* (:attr:`DiskChopper.slit_end`) angles
        measured from TDC.
        In NeXus, they are stored together as ``slit_edges``.
        :meth:`DiskChopper.from_nexus` can extract the begin and end edges from
        this combined array.
    * - ``frequency``
      - :math:`f`
      - The rotation frequency of the chopper.
        Stored in :attr:`DiskChopper.frequency`.
        A positive frequency means anticlockwise rotation and a negative frequency
        clockwise rotation (as seen from the source).
    * - ``angular_frequency``
      - :math:`\omega`
      - :math:`\omega = 2 \pi f`, :attr:`DiskChopper.angular_frequency`.
    * - ``delay``
      - :math:`\delta t`
      - Delay of the chopper timing system relative to global facility time with
        :math:`t_g = t + \delta t`, where :math:`t_g` is a global time and :math:`t`
        a chopper time.
    * - ``phase``
      - :math:`\phi`
      - The phase of the chopper relative to the pulse time.
        Defined as :math:`\phi = \omega (t_0 + \delta t - T_0)`, see below for
        the explanation.
        (:attr:`DiskChopper.phase`).
    * - ``pulse_time``
      - :math:`T_0`
      - Timestamp of a neutron pulse in global facility time.

.. image:: /_static/chopper/chopper-coordinates.svg
   :width: 400
   :align: center

:class:`DiskChopper` expects the chopper to be in phase with the source.
It thus requires a constant rotation speed which must be an integer
multiple of the source frequency or vice versa.
The phase should be computed as defined about from the difference of a pulse time
and a corresponding TDC timestamp.
The user is responsible for determining the correct times.

Slit openings
-------------

The terminology here differentiates slit 'begin' and 'end' from 'open' and 'close'.
The former refer to the angles relative to TDC as shown in the image above.
The latter refer to the times when a slit opens and closes for the beam.

It is possible to have ``end > 360 deg`` if a slit spans TDC.

For a given slit, we require ``begin < end``.
To also have ``open < close`` for both directions of rotation,
we have the following correspondence:

- clockwise rotation: ``begin`` <-> ``open`` and ``end`` <-> ``close``
- anticlockwise rotation: ``begin`` <-> ``close`` and ``end`` <-> ``open``

Time calculations
-----------------

Given the definitions above, the time in the global timing system when a point
at angle :math:`\theta` is at the beam position is

.. math::

    t_g(\theta) &= t_0 + \delta t + \begin{cases}
    \frac{\theta-\tilde{\theta}}{|\omega|}, & \textsf{clockwise}\\
    \frac{2\pi - (\theta-\tilde{\theta})}{|\omega|}, & \textsf{anticlockwise}
    \end{cases}\\
    &= t_0 + \delta t + \frac{\tilde{\theta} - \theta}{\omega} + \begin{cases}
    0, & \textsf{clockwise}\\
    \frac{2\pi}{\omega}, & \textsf{anticlockwise}
    \end{cases}

where the second line uses that, for clockwise rotation, :math:`|\omega| = -\omega`
and for anticlockwise, :math:`|\omega| = \omega`.
This can be converted to a time offset from a pulse time :math:`T_0` using

.. math::

    \Delta t_g(\theta) = t_g(\theta) - T_0 = \frac{\tilde{\theta}
       + \phi - \theta}{\omega}
       + \begin{cases}
         0, & \textsf{clockwise}\\
         \frac{2\pi}{\omega}, & \textsf{anticlockwise}
         \end{cases}

where :math:`\phi = \omega (t_0 + \delta t - T_0)` is the ``phase``.

:meth:`DiskChopper.time_offset_angle_at_beam` can calculate :math:`\Delta t_g(\theta)`
and :meth:`DiskChopper.time_offset_open` :meth:`DiskChopper.time_offset_close` calculate
:math:`\Delta t_g` for slit open and close times.

The definitions used here can lead to surprising results.
The plots below show the computed times for an angle :math:`\theta` for
:math:`\tilde{\theta} \neq 0` and :math:`\phi = 0` (blue lines).
Note in particular the time ranges for :math:`\theta \in [0, 2\pi)` (gray rectangles).
The other lines show :math:`\Delta t_g \,\mathsf{mod}\, 1/f` which is an option
for restricting the times onto :math:`[0, \frac{2\pi}{|\omega|}) = [0, \frac1{f})`.

.. image:: /_static/chopper/disk-chopper-time-curve.svg
   :class: only-light
   :width: 700
   :align: center

.. image:: /_static/chopper/disk-chopper-time-curve-dark.svg
   :class: only-dark
   :width: 700
   :align: center

|

:meth:`DiskChopper.time_offset_open` and :meth:`DiskChopper.time_offset_close`
compute :math:`\Delta t_g` for the slit edges.
The resulting times are shown below for a chopper with two slits, a short one (blue)
and a long one (orange/yellow) with
:math:`\theta_\mathsf{short} < \theta_\mathsf{long}`.
Source pulses are indicated in gray.
The shown frequency ratio is ``frequency / pulse_frequency``.

Times were computed for a pulse at :math:`T_0` with length :math:`\Delta T` but spill
into neighboring pulses because of a nonzero beam position and phase.
If those were zero, the :math:`\mathsf{frequency\ ratio} \geq 1` openings would lie
within :math:`[T_0,\, T_0 + \Delta T)` and the :math:`\mathsf{frequency\ ratio} = 0.5`
openings would lie within :math:`[T_0,\, T_0 + 2\Delta T)`.
Note how, for choppers that spin at a multiple of the pulse frequency, there are
multiple openings per slit.

.. image:: /_static/chopper/disk-chopper-openings.svg
   :class: only-light
   :width: 700
   :align: center

.. image:: /_static/chopper/disk-chopper-openings-dark.svg
   :class: only-dark
   :width: 700
   :align: center

.. _disk_chopper-time_dependent_parameters:

Time-dependent parameters
-------------------------

In NeXus files, many chopper parameters are time-dependent, for example,
``top_dead_center`` is an array of timestamps, or ``frequency`` typically
is an ``NXlog`` with speed measurements for different times.
However, for simplicity and efficiency, :class:`DiskChopper` requires
time-independent quantities.
The guide on `pre-processing choppers <../../user-guide/chopper/pre-processing.ipynb>`
shows how to extract the required quantities.
"""

from __future__ import annotations

import dataclasses
from collections.abc import Mapping
from typing import Any
from uuid import uuid4

import numpy as np
import scipp as sc
import scipp.constants

try:
    # Python 3.11+
    from enum import StrEnum

    class DiskChopperType(StrEnum):
        """Type of disk chopper."""

        single = 'Chopper type single'
        contra_rotating_pair = 'contra_rotating_pair'
        synchro_pair = 'synchro_pair'

    del StrEnum

except ImportError:
    from enum import Enum

    class DiskChopperType(str, Enum):  # type: ignore[no-redef]
        """Type of disk chopper."""

        single = 'Chopper type single'
        contra_rotating_pair = 'contra_rotating_pair'
        synchro_pair = 'synchro_pair'

    del Enum


@dataclasses.dataclass(frozen=True, eq=False)
class DiskChopper:
    """A disk chopper.

    This is a dataclass that encodes parameters of a single disk chopper and
    provides methods for computing slit opening times.

    This class requires that the chopper is in phase with the neutron source and that
    the rotation frequency is constant.
    It, therefore, does not accept time-dependent parameters.
    If your data is time-dependent, you need to select a time range where the chopper
    is in phase and compute a constant rotation frequency and phase for this range.
    In particular, the top-dead-center timestamps need to be subsumed by the ``phase``
    together with the source frequency.

    See Also
    --------
    scippneutron.chopper.disk_chopper:
        For detailed documentation of the definitions and calculations used
        by ``DiskChopper``.
    scippneutron.chopper.nexus_chopper.post_process_disk_chopper:
        A function for converting NeXus chopper data into a supported layout.
    """

    axle_position: sc.Variable
    """Position of the chopper.

    This is the center point of the chopper's axle in the face towards the source.
    See https://manual.nexusformat.org/classes/base_classes/NXdisk_chopper.html
    """
    frequency: sc.Variable
    """Rotation frequency of the chopper."""
    beam_position: sc.Variable
    """Angle where the beam crosses the chopper."""
    phase: sc.Variable
    r"""Phase of the chopper rotation relative to the source pulses.

    Defined as :math:`\phi = \omega (t_0 + \delta t - T_0)`, where :math:`t_0` is a
    TDC timestamp, :math:`\delta t` is the chopper delay, and  :math:`T_0`
    is the pulse time.
    """
    slit_begin: sc.Variable
    """Begin-edges of the slits as angles measured anticlockwise from top-dead-center.

    The order is arbitrary but must match the order of ``slit_end``.
    """
    slit_end: sc.Variable
    """Begin-edges of the slits as angles measured anticlockwise from top-dead-center.

    The order is arbitrary but must match the order of ``slit_end``.
    """
    slit_height: sc.Variable | None = None
    """Distance from chopper outer edge to bottom of slits."""
    radius: sc.Variable | None = None
    """Radius of the chopper."""

    def __post_init__(self) -> None:
        # Check for frequency because not all NeXus files store a unit
        # and the name can be confusing.
        _require_frequency('frequency', self.frequency)
        _check_edges(self.slit_begin, self.slit_end)
        object.__setattr__(
            self,
            'slit_height',
            _broadcast_slit_height(self.slit_height, self.slit_begin),
        )

    @classmethod
    def from_nexus(
        cls, chopper: Mapping[str, sc.Variable | sc.DataArray]
    ) -> DiskChopper:
        """Construct a new DiskChopper from data loaded from NeXus.

        Keys in the input correspond to the fields of `NXdisk_chopper
        <https://manual.nexusformat.org/classes/base_classes/NXdisk_chopper.html>`_.
        The values have to be post-processed, e.g., with
        :func:`~scippneutron.chopper.nexus_chopper.post_process_disk_chopper`.
        See its documentation for the required steps.
        Also, note the class docs of :class:`DiskChopper`
        about time-dependent fields.

        Parameters
        ----------
        chopper:
            Data group with fields loaded from an ``NXdisk_chopper``
            and post-processed as described above.

        Returns
        -------
        :
            A new ``DiskChopper`` instance.
        """
        if (
            typ := chopper.get('type', DiskChopperType.single)
        ) != DiskChopperType.single:
            raise NotImplementedError(
                'Class DiskChopper only supports single choppers,'
                f'got chopper type {typ}'
            )
        return DiskChopper(
            axle_position=chopper['position'],
            frequency=_get_1d_variable(chopper, 'rotation_speed'),
            beam_position=_get_1d_variable(chopper, 'beam_position'),
            phase=_get_1d_variable(chopper, 'phase'),
            slit_height=chopper.get('slit_height'),
            radius=chopper.get('radius'),
            **_get_edges_from_nexus(chopper),
        )

    @property
    def n_slits(self) -> int:
        """Number of slits."""
        if self.slit_begin.ndim != 1:
            raise sc.DimensionError(
                'Cannot determine the number of slits because '
                'the edges are not a 1d array.'
            )
        return len(self.slit_begin)

    @property
    def angular_frequency(self) -> sc.Variable:
        """Rotation speed as an angular frequency in ``rad * frequency.unit``."""
        return sc.scalar(2.0, unit="rad") * sc.constants.pi * self.frequency

    @property
    def is_clockwise(self) -> bool:
        """Return True if the chopper rotates clockwise."""
        return (self.frequency < 0.0 * self.frequency.unit).value

    def time_offset_open(self, *, pulse_frequency: sc.Variable) -> sc.Variable:
        r"""Return the opening time offsets of the chopper slits.

        Computes :math:`\Delta t_g(\theta)` as defined in
        :mod:`scippneutron.chopper.disk_chopper` with :math:`\theta` = ``slit_begin``
        for clockwise rotation and :math:`\theta` = ``slit_end`` otherwise.

        If the chopper spins at a multiple of the pulse frequency, each slit shows up
        multiple times in the result such that the array covers an entire pulse
        length.
        If the chopper spins slower than the pulse frequency, only one time per slit
        is returned, but the result covers more than one pulse in time.
        See the module documentation :mod:`scippneutron.chopper.disk_chopper`
        for details.

        Parameters
        ----------
        pulse_frequency;
            Frequency of the neutron source.

        Returns
        -------
        :
            Variable of opening times as offsets from the pulse time.
        """
        slit_edges = self.slit_begin if self.is_clockwise else self.slit_end
        return self.time_offset_angle_at_beam(
            angle=slit_edges, n_repetitions=self._source_phase_factor(pulse_frequency)
        )

    def time_offset_close(self, *, pulse_frequency: sc.Variable) -> sc.Variable:
        r"""Return the opening time offsets of the chopper slits.

        Computes :math:`\Delta t_g(\theta)` as defined in
        :mod:`scippneutron.chopper.disk_chopper` with :math:`\theta` = ``slit_end``
        for clockwise rotation and :math:`\theta` = ``slit_start`` otherwise.

        If the chopper spins at a multiple of the pulse frequency, each slit shows up
        multiple times in the result such that the array covers an entire pulse
        length.
        If the chopper spins slower than the pulse frequency, only one time per slit
        is returned, but the result covers more than one pulse in time.
        See the module documentation :mod:`scippneutron.chopper.disk_chopper`
        for details.

        Parameters
        ----------
        pulse_frequency;
            Frequency of the neutron source.

        Returns
        -------
        :
            Variable of opening times as offsets from the pulse time.
        """
        slit_edges = self.slit_end if self.is_clockwise else self.slit_begin
        return self.time_offset_angle_at_beam(
            angle=slit_edges, n_repetitions=self._source_phase_factor(pulse_frequency)
        )

    def time_offset_angle_at_beam(
        self, *, angle: sc.Variable, n_repetitions: int = 1
    ) -> sc.Variable:
        r"""Return the time offset when an angle on the chopper is at the beam.

        The time is an offset from the given pulse time.
        It encodes the time when the given angle passes by the beam position.

        Computes :math:`\Delta t_g(\theta)` as defined in
        :mod:`scippneutron.chopper.disk_chopper`.

        Parameters
        ----------
        angle:
            Angle to compute time for.
            Defined anticlockwise with respect to top-dead-center.
        n_repetitions:
            How many chopper rotations to perform.

        Returns
        -------
        :
            Computed time offset.
        """
        angle = self._apply_angle_repetitions(angle=angle, n_repetitions=n_repetitions)
        angle = (
            self.beam_position.to(unit='rad', dtype='float64')
            + self.phase.to(unit='rad', dtype='float64')
            - angle.to(unit='rad', dtype='float64', copy=False)
        )
        if not self.is_clockwise:
            angle = sc.scalar(2.0, unit='rad') * sc.constants.pi + angle
        return angle / self.angular_frequency

    def open_duration(self, *, pulse_frequency: sc.Variable) -> sc.Variable:
        """Return how long the chopper is open for.

        Parameters
        ----------
        pulse_frequency;
            Frequency of the neutron source.

        Returns
        -------
        :
            Variable of opening durations.
        """
        return self.time_offset_close(
            pulse_frequency=pulse_frequency
        ) - self.time_offset_open(pulse_frequency=pulse_frequency)

    def __eq__(self, other: object) -> bool | NotImplemented:
        if not isinstance(other, DiskChopper):
            return NotImplemented
        return all(
            _field_eq(getattr(self, field.name), getattr(other, field.name))
            for field in dataclasses.fields(self)
        )

    def make_svg(self, image_size: int = 400) -> str:
        """Generate an SVG image for this chopper.

        Parameters
        ----------
        image_size:
            The size in pixels of the image.

        Returns
        -------
        :
            The SVG image as a string.
        """
        from ._svg import draw_disk_chopper

        return draw_disk_chopper(self, image_size=image_size)

    def _repr_svg_(self) -> str:
        return self.make_svg()

    def _repr_html_(self) -> str:
        from .._html_repr import disk_chopper_html_repr

        return disk_chopper_html_repr(self)

    def _apply_angle_repetitions(
        self, *, angle: sc.Variable, n_repetitions: int
    ) -> sc.Variable:
        dim = str(uuid4())
        # Start at -1 to ensure a rotation that is finishing when the pulse begins is
        # also included.
        repetition_offsets = sc.arange(dim, -1, n_repetitions, unit='rad') * (2 * np.pi)
        if self.is_clockwise:
            repeated = angle + repetition_offsets.to(unit=angle.unit)
        else:
            # Make sure the repeated angles are later in time.
            repeated = angle - repetition_offsets.to(unit=angle.unit)

        if angle.ndim == 0:
            return repeated.flatten(to='slit')
        # Remove aux dimension.
        return repeated.transpose([*angle.dims[:-1], dim, angle.dims[-1]]).flatten(
            dims=[dim, angle.dims[-1]], to=angle.dims[-1]
        )

    def _source_phase_factor(self, pulse_frequency: sc.Variable) -> int:
        if self.frequency.ndim != 0:
            raise sc.DimensionError(
                'The chopper rotation speed must be a scalar, '
                f'got dims {self.frequency.sizes}.'
            )
        if pulse_frequency.ndim != 0:
            raise sc.DimensionError(
                'The pulse frequency must be a scalar, '
                f'got dims {pulse_frequency.sizes}.'
            )
        if pulse_frequency.value <= 0:
            raise ValueError(
                f'The pulse frequency must be > 0, got {pulse_frequency:c}.'
            )

        frequency = abs(self.frequency)
        pulse_frequency = pulse_frequency.to(unit=frequency.unit)
        quot = frequency / pulse_frequency
        if not _is_int_or_inverse_int(quot, rtol=sc.scalar(1e-8)):
            raise ValueError(
                'The chopper is out of phase with the source. '
                'The frequency must be an integer multiple of the '
                'pulse frequency or vice versa.\n'
                f'pulse_frequency:\n  {pulse_frequency}\n'
                f'frequency:\n  {self.frequency}'
            )
        # If pulse_frequency > frequency, quot < 0 but we want 1 repetition
        # of the slits, so use `max` here:
        return round(max(quot.value, 1))


def _field_eq(a: Any, b: Any) -> bool:
    if isinstance(a, sc.Variable | sc.DataArray):
        try:
            return sc.identical(a, b)
        except TypeError:
            return False  # if identical does not support b
    return a == b


def _require_frequency(name: str, x: sc.Variable) -> None:
    try:
        sc.scalar(0.0, unit=x.unit).to(unit='Hz')
    except sc.UnitError:
        raise sc.UnitError(f"'{name}' must be a frequency, got unit {x.unit}") from None


def _check_edges(begin: sc.Variable, end: sc.Variable) -> None:
    if begin.sizes != end.sizes:
        raise sc.DimensionError(
            'The begin and end edges have different sizes: '
            f'{begin.sizes} and {end.sizes}'
        )
    if sc.any(begin > end):
        raise ValueError('The begin edges must be smaller than the end edges.')
    _check_edge_overlap(begin, end)


def _check_edge_overlap(begin: sc.Variable, end: sc.Variable) -> None:
    edges = sc.concat([begin.flatten(to='slit'), end.flatten(to='slit')], dim='edge')
    edges = sc.sort(edges, key=edges['edge', 0])
    begin, end = edges['edge', 0], edges['edge', 1]
    if sc.any(begin[1:] <= end[:-1]):
        raise ValueError('The chopper has overlapping slits.')


def _broadcast_slit_height(
    slit_height: sc.Variable | None, slit_begin: sc.Variable
) -> sc.Variable | None:
    if slit_height is None:
        return None
    return slit_height.broadcast(sizes=slit_begin.sizes)


def _get_edges_from_nexus(
    nexus_chopper: Mapping[str, sc.Variable | sc.DataArray],
) -> dict[str, sc.Variable]:
    if (edges := nexus_chopper.get('slit_edges')) is not None:
        if 'slit_begin' in nexus_chopper or 'slit_end' in nexus_chopper:
            raise ValueError(
                'NeXus chopper data contains both slit_edges and slit_begin or slit_end'
            )
        if edges.ndim != 1:
            raise sc.DimensionError("The slit edges must be 1-dimensional")

        begin = edges[::2]
        end = edges[1::2]
        if sc.any(begin > end):
            raise ValueError(
                "Invalid slit edges, must be given as "
                "[begin_0, end_0, begin_1, end_1, ...] where begin_n < end_n"
            )
        return {'slit_begin': begin, 'slit_end': end}

    return {
        'slit_begin': nexus_chopper['slit_begin'],
        'slit_end': nexus_chopper['slit_end'],
    }


def _len_or_1(x: sc.Variable) -> int:
    if x.ndim == 0:
        return 1
    return len(x)


def _get_1d_variable(
    dg: Mapping[str, sc.Variable | sc.DataArray], name: str
) -> sc.Variable:
    if (val := dg.get(name)) is None:
        raise ValueError(f"Chopper field '{name}' is missing")

    msg = (
        "Chopper field '{name}' must be a scalar variable, {got}. "
        "See the chopper user-guide for more information: "
        "https://scipp.github.io/scippneutron/user-guide/chopper/pre-processing.html"
    )

    if not isinstance(val, sc.Variable):
        raise TypeError(msg.format(name=name, got=f'got a {type(val)}'))
    if val.ndim != 0:
        raise sc.DimensionError(
            msg.format(name=name, got=f'got a {val.ndim}d variable')
        )
    return val


def _is_int_or_inverse_int(x: sc.Variable, *, rtol: sc.Variable) -> bool:
    a = sc.all(abs(sc.round(x) - x) < rtol)
    y = sc.reciprocal(x)
    b = sc.all(abs(sc.round(y) - y) < rtol)
    return bool(a | b)



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)
# @author Jan-Lukas Wynen

import pytest
import scipp as sc

import scippneutron as scn

try:
    import mantid.kernel as kernel
    import mantid.simpleapi as sapi
except ImportError:
    pytestmark = pytest.mark.skip('Mantid framework is unavailable')
    sapi = None
    kernel = None


@pytest.fixture
def in_ws():
    ws = sapi.CreateSampleWorkspace(
        SourceDistanceFromSample=10.0,
        BankDistanceFromSample=1.1,
        BankPixelWidth=2,
        NumBanks=1,
        XMax=200,
        StoreInADS=False,
    )
    return ws


@pytest.fixture
def in_da(in_ws):
    return scn.mantid.from_mantid(in_ws)


def test_beamline_compute_l1(in_ws, in_da):
    out_mantid = in_ws.detectorInfo().l1() * sc.Unit('m')
    in_da = scn.mantid.from_mantid(in_ws)['data']
    out_scipp = scn.L1(in_da)
    assert sc.allclose(
        out_scipp, out_mantid, rtol=1e-15 * sc.units.one, atol=1e-15 * out_scipp.unit
    )


def test_beamline_compute_l2(in_ws, in_da):
    out_mantid = sc.array(
        dims=['spectrum'],
        unit='m',
        values=[in_ws.detectorInfo().l2(i) for i in range(in_ws.detectorInfo().size())],
    )
    in_da = scn.mantid.from_mantid(in_ws)['data']
    out_scipp = scn.L2(in_da)
    assert sc.allclose(
        out_scipp, out_mantid, rtol=1e-15 * sc.units.one, atol=1e-15 * out_scipp.unit
    )


def test_beamline_compute_two_theta(in_ws, in_da):
    out_mantid = sc.array(
        dims=['spectrum'],
        unit='rad',
        values=[
            in_ws.detectorInfo().twoTheta(i) for i in range(in_ws.detectorInfo().size())
        ],
    )
    in_da = scn.mantid.from_mantid(in_ws)['data']
    out_scipp = scn.two_theta(in_da)
    assert sc.allclose(
        out_scipp, out_mantid, rtol=1e-13 * sc.units.one, atol=1e-13 * out_scipp.unit
    )
